/*
Copyright (c) 2001, John Lambert jlambert@jlambert.com
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are 
permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of 
conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice, this list
of conditions and the following disclaimer in the documentation and/or other
materials provided with the distribution. 

Neither the name of the xdProf project nor the names of its contributors may be used
to endorse or promote products derived from this software without specific prior 
written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
/*
$Header: /xdprof-dll/background.cpp 11    5/09/01 5:29a Admin $ 
 */
/** \file background.cpp
 * Things related to background processing and threads.
 */
#include "xdprof.h"



//! How often to refresh (in milliseconds)
int RefreshTime = 2000;

//! Information about this computer
static string SelfInfo;

//! The monitor used for background notification/shutdown
volatile JVMPI_RawMonitor BackgroundMonitor;

//! The hostname to connect to.
static string HostName;

//! The port number to connect to.
static int PortNumber;

//! set to true by external thread when the thread should shutdown
volatile bool ShutdownRequested;

static void DisableEvents()
{
CHECKEDCALL(DisableEvent(JVMPI_EVENT_JVM_INIT_DONE               , NULL));
CHECKEDCALL(DisableEvent(JVMPI_EVENT_JVM_SHUT_DOWN				, NULL));
CHECKEDCALL(DisableEvent(JVMPI_EVENT_THREAD_START, NULL));
CHECKEDCALL(DisableEvent(JVMPI_EVENT_THREAD_END, NULL));
CHECKEDCALL(DisableEvent(JVMPI_EVENT_CLASS_LOAD                  , NULL));
CHECKEDCALL(DisableEvent(JVMPI_EVENT_CLASS_UNLOAD                , NULL));
CHECKEDCALL(DisableEvent(JVMPI_EVENT_OBJECT_ALLOC                  , NULL));

}

/**
Initialize the background thread by setting the selfinfo and refreshtime.
@param host the hostname to connect to
@param port the port number to use
@param selfinfo a descriptive string about this computer/vm/connection
@param refreshTime how often to refresh in seconds
  */
void InitializeBackgroundThread(string host, int port, string selfinfo, int refreshTime)
{
	HostName = host;
	PortNumber = port;
	SelfInfo = selfinfo;
	RefreshTime = refreshTime;
}

/**
This is the background thread; it sends the self info once and then updates the server
every #RefreshTime milliseconds.
  */
extern "C" void BackgroundThread(void *)
{
	// must be created by this function
	BackgroundMonitor = CALL(RawMonitorCreate)("_background");
	CALL(RawMonitorEnter)(BackgroundMonitor);

	// now set up the communications
	bool ok = InitializeCommunications(HostName, PortNumber);
	if (!ok || ! IsConnected())
	{
		debug("couldn't connect, disabling events");
		// TODO: shut it down
		ShutdownCommunications();
		DisableEvents();
		return;
	}
	
	// Once and only once, we send the SelfInfo
	//Send(SelfInfo + "\n");
	ShutdownRequested = false;

	ThreadID me;

    (*GlobalJVM).GetEnv((void **)&me, JNI_VERSION_1_2);
	jclass c = me->FindClass("java/lang/System");
    jmethodID m = me->GetStaticMethodID(c, "getProperty", "(Ljava/lang/String;)Ljava/lang/String;"); 
    vector<string> properties;
	properties.push_back("os.arch");
	properties.push_back("os.name");
	properties.push_back("os.version");
	properties.push_back("java.vm.name");
	properties.push_back("java.vm.info");
	properties.push_back("java.vm.version");
	properties.push_back("java.vm.vendor");
	properties.push_back("java.runtime.name");
	properties.push_back("java.runtime.version");

	ostringstream oss;
	for (vector<string>::size_type  i = 0; i < properties.size(); i++)
	{
		jstring str = (jstring)me->CallStaticObjectMethod(c, m, me->NewStringUTF(properties[i].c_str()));
		//cout << properties[i] << "\t" << me->GetStringUTFChars(str, 0) << endl;
		debug(i, properties[i], me->GetStringUTFChars(str, 0));
		oss << properties[i] << "=" << me->GetStringUTFChars(str, 0) << '!';
	}
	oss << ends;
	string ost(oss.str());
	SelfInfo += " !";
	SelfInfo += ost.substr(0, ost.size() - 2) ;
	SelfInfo += '\n';
	debug("selfinfo", SelfInfo);
	debug("background thread id", me);

	while(! ShutdownRequested)
	{
		CALL(RawMonitorWait)(BackgroundMonitor, (jlong)RefreshTime);
		if (ShutdownRequested)
			break;
		if (IsConnected())
			ThreadDump(SelfInfo, me);
		else
		{
			debug("lost connection, shutting down.");
			DisableEvents();
			ShutdownCommunications();
			return;
		}
	}
	//	debug("out of shutdown loop.");
	CALL(RawMonitorDestroy)(BackgroundMonitor);
}

/**
Creates the background thread.
  */
void CreateBackgroundThread()
{
	CALL(CreateSystemThread)("DPROF Background Thread", JVMPI_NORMAL_PRIORITY, BackgroundThread);
	debug("bg thread created");
}
