/*
Copyright (c) 2001, John Lambert jlambert@jlambert.com
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are 
permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of 
conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice, this list
of conditions and the following disclaimer in the documentation and/or other
materials provided with the distribution. 

Neither the name of the xdProf project nor the names of its contributors may be used
to endorse or promote products derived from this software without specific prior 
written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
/*
$Header: /xdprof-dll/communication.cpp 13    5/09/01 5:29a Admin $ 
 */
#include "xdprof.h"

/** \file communication.cpp
 * Things related to network communication.
 */


/**
 * Socket connection (or < 0 if not connected)
 */
static int Server;

//! Number of bytes sent
static int BytesSent = 0;

//! Number of times Send was called
static int SendCalls = 0;

/**
Sends the msg of length len over the socket.
@param msg the data to send
@param len the length of the data
  */
int Send(const char* msg, int len)
{
	if (Server > 0)
	{
		BytesSent += len;
		++SendCalls;
		return send(Server, msg, len, 0); 
	}
	else
		return 0;
}

/**
 Sends a string.
 @param s the string to send
  */
void Send(const string & s)
{
	Send(s.c_str(), s.length() );
}

/**
Determines if the client is connected to the server.
*/
bool IsConnected()
{
	return (Server > 0);
}

/**
 * Return a socket descriptor connect()ed to a hostname that is
 * accept()ing information data on port. Return a value <= 0 if
 * such a connection can't be made.
 * @param hostname the hostname to connect to
 * @param port the port number
 */
static int GetSocketDescriptor(const char *hostname, unsigned short port)
{
	struct hostent *hentry;
	struct sockaddr_in s;
	int fd;
	
	if (port <= 0 || port > 65535) 
	{
		fprintf(stderr, "DPROF ERROR: bad port number\n");
		return -1;
	}
	
	if (hostname == NULL) 
	{
		fprintf(stderr, "DPROF ERROR: hostname is NULL\n");
		return -1;
	}
	
	/* create a socket */
	fd = socket(AF_INET, SOCK_STREAM, 0);
	
	/* find remote host's addr from name */
	if ((hentry = gethostbyname(hostname)) == NULL)
	{
#ifdef WIN32
		debug("Bad gethostbyname value.", WSAGetLastError());
#endif
		return -1;
	}
	
	memset((char *)&s, 0, sizeof(s));
	/* set remote host's addr; its already in network byte order */
	memcpy(&s.sin_addr.s_addr, *(hentry->h_addr_list), 
		sizeof(s.sin_addr.s_addr));
	/* set remote host's port */
	s.sin_port = htons(port);
	s.sin_family = AF_INET;
	
	/* now try connecting */
	if (-1 == connect(fd, (struct sockaddr*)&s, sizeof(s))) 
	{
		debug("Unable to connect!");
		return -1;
	}
	else 
	{
		return fd;
	}
	
}

/**
This initializes communications with the server.
@param hostname the hostname to connect to; either IP address or a machine name
@param port the port number to use
@param selfinfo a descriptive string describing this connection
  */
bool InitializeCommunications(string hostname, int port)
{
#ifdef WIN32	
	WORD wVersionRequested;
	WSADATA wsaData;
	int err;
 
	wVersionRequested = MAKEWORD( 1, 1 );
 
	err = WSAStartup( wVersionRequested, &wsaData );
	if ( err != 0 ) {
		debug("WSA error", WSAGetLastError());
		Server = -1;
		return false;
	}
#endif

	Server = GetSocketDescriptor(hostname.c_str(), static_cast<unsigned short>(port));
	if (IsConnected())
		debug("connected!");
	else
		debug("couldn't connect");
	return IsConnected();
}

static int PacketCount = 0;
static unsigned long int PacketBytes = 0;

void RecordPacket(unsigned long packetsize)
{
	++PacketCount;
	PacketBytes += packetsize;
}

/**
Shutdown the communications, close the sockets, etc.
*/
void ShutdownCommunications()
{
	Send("/*done*/");
	unsafe_debug("Bytes sent:", BytesSent);
	unsafe_debug("SendCalls: ", SendCalls);
	//fprintf(stderr, "bytes=%d calls=%d\n", BytesSent, SendCalls);
	fprintf(stdout, "\npackets\tbytes\trefresh\n");
	//fprintf(stderr, "packets\t%d \nbytes\t%d\nrefresh\t%d\n", PacketCount, PacketBytes,RefreshTime);
	fprintf(stdout, "%d\t%d\t%d\n", PacketCount, PacketBytes,RefreshTime);

fprintf(stdout, "Methods=%d, Classes=%d, MethodCache=%d,ClassCache=%d\n", Methods.size(), Classes.size(), MethodCache.size(), ClassCache.size());

	fflush(stdout);
	unsafe_debug("Average: ", BytesSent / static_cast<double>(SendCalls));
	if (Server > 0)
		closesocket(Server);
	Server = -1;
#ifdef WIN32
	WSACleanup();
#endif
}
