package org.xdprof.data;

import java.util.*;
import org.xdprof.util.Debug;

public class ThreadInfo implements Comparable 
{
	public int compareTo(Object o)
	{
		if (o instanceof ThreadInfo)
			return m_started - ((ThreadInfo)o).m_started;
		else
			throw new ClassCastException("bad cast in compare to");
	}
	private final ThreadIdentifier m_threadID;
	private final String m_name, m_group;
	private final String m_parent;
	private final int m_status;
	private final Trace m_context;
	private final int m_started;
	
	public String getParent()
	{
		return m_parent;	
	}
	
	public String getName()
	{
		return m_name;	
	}
	
	public String getGroup()
	{
		return m_group;	
	}
	
	public ThreadInfo(Trace context, ThreadIdentifier tid, String name, String group, String parent, int started, int status)
	{
		m_context = context;
		m_threadID = tid;
		m_name = name;
		m_group = group;
		m_parent = parent;
		m_status = status;
		m_started = started;
	}

	public class StackFrame
	{
		public final int getLine()
		{
			return m_line;	
		}
		
		private final String getLineNumber()
		{
			if (m_line >= 0)
				return "" + m_line;
			if (m_line == -2)
				return "Compiled method";
			if (m_line == -3)
				return "Native method";
			else //if (m_lineNumber == -1)
				return "Unknown line";	
		}
		private final MethodIdentifier m_method;
		private final int m_line;
		
		public MethodIdentifier getMethodID()
		{
			return m_method;	
		}
		
		public StackFrame(MethodIdentifier mid, int line)
		{
			m_method = mid;
			m_line = line;
		}
		
		private String m_description = null;
		MethodInfo mi;
		ClassInfo ci;
		public String toString()
		{
			if (m_description == null)
			{
			mi = m_context.findMethodInfo(m_method);
			ci = m_context.findClassInfo(mi.getClassID());	
			m_description = mi.toString() + " " + ci.toString() + ": " + getLineNumber();
			}
			return m_description;
		}
		
		public Vector toVector()
		{
			toString();
			Vector v = new Vector();
			v.add(mi.toString());
			v.add(ci.toString());
			v.add(getLineNumber());
			return v;
		}
	}
	
	private final ArrayList m_frames = new ArrayList();
	
	public int getFrameCount()
	{
		return m_frames.size();	
	}
	
	public MethodIdentifier getFrameMethod(int frameNumber)
	{
		return ((StackFrame)m_frames.get(frameNumber)).getMethodID();
	}
	
	public int getFrameLineNumber(int frameNumber)
	{
		return ((StackFrame)m_frames.get(frameNumber)).getLine();
	}
	
	/**
	 * Adds this frame to the bottom of the stack.
	 */
	public void addStackFrame(MethodIdentifier mid, int lineNumber)
	{
		//Debug.trace(m_context.toString() +  " stack adding " + mid + " " + lineNumber);
		m_frames.add(new StackFrame(mid, lineNumber) );
	}
	
	/**
	 * Returns a vector of the StackFrame's.
	 */
	public Vector getStackVector()
	{
		return new Vector(m_frames);
	}
	
	public ThreadIdentifier getThreadID()
	{
		return m_threadID;	
	}

	/**
	 *  SUSPENDED = 0x8000
	 */
	public static final int SUSPENDED = 0x8000;
	
	/**
	 * INTERRUPTED = 0x4000
	 */
	public static final int INTERRUPTED = 0x4000;
	
	/**
	 * RUNNABLE = 1
	 */
    public static final int RUNNABLE = 1;
	
	/**
	 * MONITOR_WAIT = 2
	 */
    public static final int MONITOR_WAIT = 2;
	
	/**
	 * CONDVAR_WAIT = 3
	 */
    public static final int CONDVAR_WAIT = 3;

	/**
	 * Returns true if the status is the flag.
	 * @param status from ThreadInfo.getStatus()
	 * @param flag from ThreadInfo constants
	 */
	public static final boolean isStatus(int status, int flag)
	{
		if (flag <= 3) return ((status & 7) == flag);
		return ((flag & status) != 0);
	}
		
	public int getStatus()
	{
		return m_status;	
	}
	
	private String m_toString = null;

	public String getThreadStatusDescription()
	{
		String s = "";
		if (isStatus(m_status, RUNNABLE))
			s = "Runnable";
		if (isStatus(m_status, CONDVAR_WAIT))
			s += "Condition Variable Wait";
		if (isStatus(m_status, MONITOR_WAIT))
			s += "Monitor Wait";
		if (isStatus(m_status, INTERRUPTED))
			s += "(Interrupted)";
		if (isStatus(m_status, SUSPENDED))
			s += "(Suspended)";
		return s;		
	}
	
	public int getStarted()
	{
		return m_started;
	}
	
	public String toString()
	{
		if (m_toString == null)
		{
		String s = "";
		if (isStatus(m_status, RUNNABLE))
			s = "Runnable";
		if (isStatus(m_status, CONDVAR_WAIT))
			s += "Condition Variable Wait";
		if (isStatus(m_status, MONITOR_WAIT))
			s += "Monitor Wait";
		if (isStatus(m_status, INTERRUPTED))
			s += "(Interrupted)";
		if (isStatus(m_status, SUSPENDED))
			s += "(Suspended)";
		String k = "      " + m_started;
			m_toString = k.substring(k.length() - 5) + " " + m_threadID + " " + m_name + " " + m_group + 
						 " " + m_parent + " " + m_started + " " + 
						 " (" + m_frames.size() + ")";
		}
		return m_toString;
	}	
}
