package org.xdprof.listener;

import org.xdprof.util.*;
import org.xdprof.server.*;
import org.xdprof.data.*;

import javax.swing.*;
import javax.swing.table.*;
import javax.swing.event.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;

/**
 * This is the Server GUI; it creates a server and provides an interface to interact with.
 */
public class ServerGUI extends JFrame implements ServerListener 
{
	/**
	 * The displayed list of connections.
	 */
	private final JList m_connectionList = new JList();
	
	/**
	 * The displayed list of threads for the current connection.
	 */
	private final JList m_threadList = new JList();
	
	/**
	 * The stack frame listing for the current thread.
	 */
	private final JList m_stackList  = new JList();
	
	/**
	 * The current connections.
	 */
	private final Vector m_connections = new Vector();
	
	/**
	 * Connections to remove after we unpause.
	 */
	private final Vector m_connectionsToRemove = new Vector();
	
	/**
	 * A HashMap from a connection object to a vector of Trace objects.
	 */
	private final HashMap m_connectionToTraceVector = new HashMap();
	
	/**
	 * The main display panel.
	 */
	private final JPanel m_pane = new JPanel(true);
	
	private boolean m_isPaused = false;
	
	/**
	 * The button for pausing.
	 */
	private final JButton m_pauseButton = new JButton("Pause");

	/**
	 * An empty vector; used to empty out the lists.
	 */
	private static final Vector EMPTY_VECTOR = new Vector();
	
	public void setPort(int p)
	{
		this.setTitle("xdProf Server on port #" + p);
	}
	
	/**
	 * The constructor.
	 * @param port the port number to create the server on
	 */
	public ServerGUI()
	{
		super("xdProf Server");
		//final Server s = new Server(port, this);
		//s.start();

		this.addWindowListener(new WindowAdapter()
			{
			public void windowClosing(WindowEvent e)
			{
				Debug.trace(e.toString());
				//s.shutdown();
				System.exit(0);
			}
	    });
		
		m_pane.setLayout(new GridLayout(2,1));
		JPanel temp = new JPanel(new GridLayout(1,2), true);
		temp.add(new JScrollPane(m_connectionList));
		m_threadList.setCellRenderer(new ThreadCellRenderer());
		temp.add(new JScrollPane(m_threadList));
		
		
		// setup the lists
		m_connectionList.addMouseListener(new MouseAdapter()
			{
			public void mouseClicked(MouseEvent e)
			{
				int index = m_connectionList.locationToIndex(e.getPoint());
				Debug.trace("Index", index);
				m_connectionList.setSelectedIndex(index);
				updateThreads();
			}
			});
 		
		m_threadList.addMouseListener(new MouseAdapter()
			{
			public void mouseClicked(MouseEvent e)
			{
				updateStack();
			}
			});
		
		m_pauseButton.addActionListener(new ActionListener()
			{
			public void actionPerformed(ActionEvent ae)
			{
				if ("Pause".equals(m_pauseButton.getText()))
				{
					m_pauseButton.setText("Unpause");
					m_isPaused = true;
				}
				else // unpause
				{
					m_pauseButton.setText("Pause");
					m_isPaused = false;
					for (Iterator i = m_connectionsToRemove.iterator(); i.hasNext(); )
					{
						removeConnection((Connection)i.next());	
					}
					m_connectionsToRemove.removeAllElements();
				}
			}
			});

		m_pane.add(temp);
		final Font f = new Font("Monospaced", Font.PLAIN, 12);
		m_stackList.setFont(f);
		m_threadList.setFont(f);
		m_pane.add(new JScrollPane(m_stackList));
        m_pane.setBorder(BorderFactory.createEmptyBorder(15, 15, 15, 15));
		m_pane.setPreferredSize(new Dimension(700,300));
		this.getContentPane().setLayout(new BorderLayout());
		this.getContentPane().add(m_pane, BorderLayout.CENTER );
		this.getContentPane().add(m_pauseButton, BorderLayout.SOUTH);
		this.setVisible(true);
		this.pack();
	}
	
	/**
	 * Render the thread list appropriately.
	 */
	private class ThreadCellRenderer extends JLabel implements ListCellRenderer
	{
		/**
		 * Formats a value.
		 * @param list the list
		 * @param value the value; should be instanceof Trace.ThreadInfo
		 * @param index position in list
		 * @param isSelected whether or not the item is selected
		 * @param cellHasFocus the list and the cell have the focus
		 */
		public Component getListCellRendererComponent(JList list, Object value, int index, boolean isSelected, boolean cellHasFocus)
		{
			String s = value.toString();
			setText(s);
			Assert.assert("", value instanceof ThreadInfo);
			
			Assert.assert("1",ThreadInfo.isStatus(ThreadInfo.RUNNABLE, ThreadInfo.RUNNABLE));
			Assert.assert("2",!ThreadInfo.isStatus(ThreadInfo.CONDVAR_WAIT, ThreadInfo.RUNNABLE));
			
			setBackground(list.getBackground());
			final int status =((ThreadInfo)value).getStatus();
			if (ThreadInfo.isStatus(status, ThreadInfo.CONDVAR_WAIT) ||  
				ThreadInfo.isStatus(status, ThreadInfo.MONITOR_WAIT) ) 
			{
				setForeground(Color.yellow.darker());
			}
			else if (ThreadInfo.isStatus(status, ThreadInfo.RUNNABLE))
			{
				setForeground(Color.green.darker().darker() );	
			}
			else
				setForeground(list.getForeground());
			if (isSelected)
				setFont(list.getFont().deriveFont(Font.BOLD) );
			else
				setFont(list.getFont());
			setEnabled(list.isEnabled());
			
			return this;
		}
	} // end ThreadCellRenderer


	
	/**
	 * Update the thread display; called when a connection is clicked on.
	 */
	private void updateThreads()
	{
		int index = m_connectionList.getSelectedIndex();
		if (index < 0)
			return;
		Debug.trace("Currently selected: connection #", index);
		Connection c = (Connection)m_connectionList.getModel().getElementAt(index);
		Debug.trace("which was: ", c);
		Vector traces = (Vector)(m_connectionToTraceVector.get(c));
		if ((traces == null) || (traces.size() <= 0))
		{
			m_threadList.setListData(EMPTY_VECTOR);//empty!
			return;
		}
		Trace t = (Trace)(traces.lastElement());
		m_threadList.setListData(t.getThreads());
	}
	
	/**
	 * Updates the stack display.
	 */
	private void updateStack()
	{
		int index = m_threadList.getSelectedIndex();
		if (index < 0)
			return;
		if (index > m_threadList.getModel().getSize())
			return;			
		Debug.trace("Double clicked on Item", index);
		ThreadInfo tt = (ThreadInfo)(m_threadList.getModel().getElementAt(index));
		Debug.trace("which was: ", tt);
		
		m_stackList.setListData(tt.getStackVector());
	}
	
	/**
	 * Adds a connection to the server.
	 * @param c the connection to register
	 */
	public void addConnection(Connection c)
	{
		m_connections.add(c);
		m_connectionToTraceVector.put(c, new Vector());
		
		if (!m_isPaused)
			m_connectionList.setListData(m_connections);
	}
	
	/**
	 * Adds a trace for the connection.
	 * @param c the connection
	 * @param t the "latest" trace
	 */
	public void addTrace(Connection c, Trace t)
	{
		Debug.trace("Adding trace for " + c);
		final Vector traces = (Vector)m_connectionToTraceVector.get(c);

		if (traces == null) // avoid race condition during disconnect
			return;
		
		traces.add(t);
		
		if (m_isPaused)
			return;
		
		if (c == m_connectionList.getSelectedValue())
		{
			ThreadInfo currentThread = (ThreadInfo) m_threadList.getSelectedValue();
			if (currentThread == null)
			{
				Debug.trace("Currentthread==null");
				updateThreads();
				return;
			}
			// if the thread they were looking at is STILL around, just update its stack
			
			Debug.trace("currentThread = ", currentThread);
			Debug.trace("currentThread.getThreadID()", currentThread.getThreadID());
			if (t.containsThread(currentThread.getThreadID()))
			{
				Debug.trace("Refreshing!!!");
				updateThreads();
				final Vector v = t.getThreads();
				for (int i = 0; i < v.size(); i++)
				{
					if ( ((ThreadInfo)v.elementAt(i)).getThreadID().equals(currentThread.getThreadID()))
					{
						m_threadList.setSelectedIndex(i);
						break;
					}
				}
				updateStack();
			}
			else
			{
				updateThreads();
			}
		}
		else
		{
		}
	}
	
	/**
	 * Removes/unregisters a connection.
	 * @param c the connection to remove
	 */
	public void removeConnection(Connection c)
	{
		if (m_isPaused)
		{
			m_connectionsToRemove.add(c);
			return;
		}
		
		boolean b = m_connections.remove(c);
		Assert.assert("removed!", b);
		Object o = m_connectionToTraceVector.remove(c);
		Assert.notNull(o);
		int oldIndex = m_connectionList.getSelectedIndex();
		m_connectionList.setListData(m_connections);
		
		if (m_connections.size() == 0 || c == m_connectionList.getSelectedValue())
		{
			m_threadList.setListData(EMPTY_VECTOR);
			m_stackList.setListData(EMPTY_VECTOR);
			return;
		}
		if (oldIndex >= 1)	
			m_connectionList.setSelectedIndex(oldIndex - 1); // select the one before it?
	}
	
	/**
	 * Start the program.
	 * @param args args[0] is the port number
	 */
	public static void main(String[] args)
	{
		try
		{
			Server.startServer(Integer.parseInt(args[0]), new ServerGUI());
		}
		catch (Exception e)
		{
			Debug.traceException("main", e);
			System.err.println("Usage: ServerGUI port");
		}
	}
}
