package org.xdprof.server;

import java.io.*;
import java.net.*;
import java.util.*;

import org.xdprof.util.*;
import org.xdprof.data.*;
import org.xdprof.listener.*;

/**
 * A connection to a client.
 */
public class ConnectionThread extends Thread
{
	/**
	 * The client's socket.
	 */
	private final Socket m_client;
	
	/**
	 * The client's hostname.
	 */
	private final String m_host;
	
	/**
	 * Number of traces so far.
	 */
	private int m_numTraces = 0;
	
	/**
	 * Descriptive string about the client.
	 */
	private String m_info;
	
	/**
	 * The listener to notify.
	 */
	private final ServerListener m_sl;
	
	/**
	 * Initializer; must be .start()'ed though.
	 * @param s socket to connect to
	 * @param sl the listener to notify
	 */
	public ConnectionThread(Socket s, ServerListener sl)
	{
		m_client = s;
		m_host = m_client.toString();
		m_sl = sl;
	}
	
	/**
	 * Returns the next non-comment line from in.
	 * @param in the BufferedReader to read from
	 */
	private String getNextLine(BufferedReader in) throws IOException 
	{
		String s = in.readLine().trim();
		while (s.startsWith("//") || s.length() == 0 )
		{
			//Debug.trace("Ignoring ", s);
			s = in.readLine().trim();
		}
//		Debug.trace("Received", s);
		Assert.assert("Length", s.length() > 0);
		Assert.assert("Starts" + s, !s.startsWith("//"));
		return s;
	}

	/**
	 * Returns an integer
	 * @param s the string containing an integer
	 */
	private static final int toInt(final String s) throws NumberFormatException
	{
		return Integer.parseInt(s.trim());	
	}
	
	/**
	 * Unmangles a string in ascii-code format by converting it to the 
	 * characters represented by the ascii-codes.
	 * @param s the mangled string
	 */
	static final String unmangle(final String s) 
	{
		try
		{
			int i = 0;
			int space = s.indexOf(' ');
			final int last = s.length();
			final StringBuffer sb = new StringBuffer(last/2);
			while (i < last)
			{
				space = s.indexOf(' ', i);
				if (space == -1) space = last;
				sb.append((char)Integer.parseInt(s.substring(i, space)));
				i = space + 1;
			}
			return sb.toString();
		}
		catch (NumberFormatException nfe)
		{
			Debug.traceException("unmangle " + s, nfe);
			throw nfe;
		}
	}
	
	/**
	 * Run the ConnectionThread
	 */
	public void run()
	{
		Connection conn = null;
		try
		{
			final BufferedReader in = new BufferedReader(new InputStreamReader(m_client.getInputStream()));
			String line = "";
			boolean alreadyRead = true;
			// the first line we read from the server will contain info about this connection
			//m_info = getNextLine(in);
			//Debug.trace("Selfinfo = ", m_info);
			m_info = line = getNextLine(in);
			
			//System.err.println("Selfinfo = " + m_info);
			
			conn = new Connection(m_info, m_host, m_client.getInetAddress(), m_client.getLocalPort());
			final Connection c = conn;
			m_sl.addConnection(c);
	
			while (line != null)
			{
				if (!alreadyRead)
					line = getNextLine(in); // self info
				else
					alreadyRead = false;
									
				if ("/*done*/".equals(line))
				{
					m_client.close();
					//Debug.trace("All done; " + m_host + " did " + m_numTraces + " traces.");
					return;
				}
				m_numTraces++;
				final Trace t = new Trace(line);		
				
				line = getNextLine(in);
				
				int numThreads = toInt(line);
				for (int i = 0; i < numThreads; i++)
				{
					final String threadID, name, group, parent;
					final int status, started;
					threadID = getNextLine(in);
					name = (getNextLine(in));
					group = (getNextLine(in));
					parent = (getNextLine(in));
					started = toInt(getNextLine(in));
					status = toInt(getNextLine(in)); // TODO: convert to description
					
					//ThreadTrace tt = t.addThread(threadID, name, group, parent, status);
					ThreadInfo ti = t.addThread(new ThreadIdentifier(threadID), name, group, parent, started, status);
										
					int numFrames = toInt(getNextLine(in));
					for (int j = 0; j < numFrames; j++)
					{
						String methodID;
						int lineNumber;
						
						line = getNextLine(in);
						int temp = line.indexOf(' ');
						int oldtemp;

						methodID = line.substring(0, temp);
						oldtemp = temp;
						lineNumber = toInt(line.substring(oldtemp + 1));
						ti.addStackFrame(new MethodIdentifier(methodID), lineNumber);
						// tt.addStackFrame(methodID, className, sourceName, methodName, methodSignature, lineNumber);						
					} // end frame
				} // end threads
				
				// now get the methods
				//Debug.trace("Reading methods");
				int numMethods = toInt(getNextLine(in));
				
				for (int i = 0; i < numMethods; i++)
				{
					//classid methodid methodname methodsig
					line = getNextLine(in);
					int temp = line.indexOf(' ');
					int oldtemp = -1;

					final String classID = line.substring(0, temp); // TODO: error!
					oldtemp = temp;
					temp = line.indexOf(' ', temp + 1);
					final String methodID = line.substring(oldtemp + 1, temp);
					oldtemp = temp;
					temp = line.indexOf(' ', temp + 1);
					final String methodName = line.substring(oldtemp + 1, temp);
					final String methodSig = line.substring(temp + 1);
							
					t.addMethod(new ClassIdentifier(classID), new MethodIdentifier(methodID), methodName, methodSig);
				}
				
				// now get classes
				//Debug.trace("Reading classes");
				int numClasses = toInt(getNextLine(in));
				//Debug.trace("To read", numClasses);
				for (int i = 0; i < numClasses; i++)
				{
					line = getNextLine(in);
					int temp = line.indexOf(' ');
					int oldtemp;

					final String classID = line.substring(0, temp);
					oldtemp = temp;
					temp = line.indexOf(' ', temp + 1);
					final String className = line.substring(oldtemp + 1, temp);
					oldtemp = temp;
					final String classSource = line.substring(oldtemp + 1);
					
					t.addClass(new ClassIdentifier(classID), className, classSource);
				}
				line = "";
				//Debug.trace("Done with classes!");
t.selfCheck();
				// so we don't block (updating GUI could take a while, etc.)
				new Thread() {
					public void run()
						{
							m_sl.addTrace(c, t);
						}
					}.start();
			} // end while
			
		}
		catch (Exception e)
		{
			Debug.traceException("ConnectionThread " + m_client.toString(), e);
		}
		finally
		{
			try
			{
				m_client.close();
			}
			catch (Exception e)
			{
				// do nothing
				Debug.traceException("inside finally block", e);
			}

			//m_sl.removeConnection(conn);
			removeConnectionThread(conn);
			System.gc(); // request a garbage collection
		}
		
	}
	
	public void removeConnectionThread(final Connection c)
	{
		new Thread()
			{
			public void run()
			{
				m_sl.removeConnection(c);
			}
			}.start();
	}
}
