package org.xdprof.util;

/**
 * Provides run-time assertion semantics for safety, etc.
 */
public class Assert
{
	/**
	 * Protected constructor to prevent instantiation
	 */
	protected Assert()
	{
	}
	
	/**
	 * Asserts that an object is not null.
	 * @param o object that shouldn't be null
	 */
	public static void notNull(Object o)
	{
		notNull("Null object", o);
	}
	
	/**
	 * Asserts that an object is not null.
	 * @param s descriptive message to display
	 * @param o object that shouldn't be null
	 */
	public static void notNull(String s, Object o)
	{
		assert(s, o != null);
	}
	
	/**
	 * Used for commenting out large blocks of debug code.
	 * @return true if Asserts/debug code should be on.
	 */
	public static boolean isOn()
	{
		return false;
	}
	
	/**
	 * Asserts that a condition is true. If it isn't it throws
	 * an AssertionFailedError with the given message.
	 * @param message the message to display if condition is false
	 * @param condition the condition to check
	 */
	public static void assert(String message, boolean condition)
	{
		if (!condition)
			fail(message);
	}
	
	/**
	 * Asserts that two objects are equal. If they are not
	 * an AssertionFailedError is thrown.
	 * @param message the detail message for this assertion
	 * @param expected the expected value of an object
	 * @param actual the actual value of an object
	 */
	public static void assertEquals(String message, Object expected, Object actual)
	{
		if (expected == null && actual == null)
			return;
		
		if (expected != null && expected.equals(actual))
			return;

		failNotEquals(message, expected, actual);
	}

	/**
	 * Asserts that two numbers are equal.
	 * @param message the message to display
	 * @param actual the value actually returned
	 * @param expected the expected value
	 */
	public static void assertEquals(String message, int actual, int expected)
	{
		if (expected == actual)
			return;

		failNotEquals(message, "" + expected, "" + actual);
	}

	/**
	 * Fails with the given message.
	 * @param message the message to display 
	 */
	public static void fail(String message)
	{
		throw new AssertionFailedError(message);
	}
	
	/**
	 * Fail with a "not equals" message
	 */
	private static void failNotEquals(String message, Object expected, Object actual)
	{
		fail(message + " expected:<" + expected + "> but was: <" + actual + ">");
	}

	/**
	 * Fail with a "not reachable" message
	 * @param s location
	 */
	public static void failNotReachable(String s)
	{
		fail("Not reachable " + s);	
	}
	
	/**
	 * Fail because something isn't reachable.
	 */
	public static void failNotReachable()
	{
		fail("Not reachable");	
	}

	/**
	 * Fail because something isn't implemented.
	 * @param s location
	 */
	public static void failNotImplemented(String s)
	{
		fail("Not Implemented " + s);	
	}
	
	/**
	 * Fail because something isn't implemented.
	 */
	public static void failNotImplemented()
	{
		fail("Not Implemented");	
	}

}