/*
Copyright (c) 2001, John Lambert jlambert@jlambert.com
All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are 
permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of 
conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice, this list
of conditions and the following disclaimer in the documentation and/or other
materials provided with the distribution. 

Neither the name of the xdProf project nor the names of its contributors may be used
to endorse or promote products derived from this software without specific prior 
written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/*
$Header: /xdprof-dll/xdprof.cpp 26    5/09/01 5:29a Admin $ 
 */
#include "xdprof.h"

//! Current methods
MethodMap Methods;

//! Currently running threads
ThreadMap Threads;

//! Class to a set of methods
ClassMethopMap ClassMethods;

//! Class to ClassInfo
ClassInfoMap Classes;

//! This is a cache of methods that have popped up in the call stack.
// TODO: figure out emptying them out
MethodMap MethodCache;
ClassInfoMap ClassCache;


//! The dead threads
static set<ThreadID> DeadThreads;

//! 


//! Whether or not we've caught the VM initialization
static bool CaughtVM = false;

static unsigned long threadcount = 0;

/**
Thread start event handler, stores information in #Threads.
 */
inline void ThreadStartHandler(ThreadID thread_id, const ThreadInfo & ti)
{
	Lock(threads_lock);
	ti.setStarted(++threadcount);
	Threads[thread_id] = ti;
	Unlock(threads_lock);
}

static void DeleteMethods()
{
	for (MethodMap::iterator mi = Methods.begin(); mi != Methods.end(); ++mi)
		delete mi->second;
}

/**
Thread end handler, removes thread appropriately.
 */
inline void ThreadEndHandler(JNIEnv *thread_id)
{
	Lock(threads_lock);
	///	jobjectID o = CALL(GetThreadObject)(thread_id);
	debug("thread end", thread_id , Threads.size());
	int k = Threads.erase(thread_id);
	DeadThreads.insert(thread_id);
	assert(k == 1);
	debug("remove ", thread_id, k);
	Unlock(threads_lock);
}

/**
 Function for handling event notification. The space allocated for the JVMPI_Event structure 
 and any event specific information is freed by the virtual machine once this function returns.
 The profiler agent must copy any necessary data it needs to retain into its internal buffers.
 @param event a point to the event
 */
void NotifyEvent(JVMPI_Event *event)
{
	switch(event->event_type)
	{
	case JVMPI_EVENT_CLASS_LOAD: 
	case JVMPI_EVENT_CLASS_LOAD | JVMPI_REQUESTED_EVENT:
		Lock(methods_lock);
		ClassLoadEvent(event->env_id, 
			        const_cast<char**>(& event->u.class_load.class_name),
			       &event->u.class_load.source_name,
			       event->u.class_load.num_interfaces,
			       event->u.class_load.num_static_fields,
			       event->u.class_load.statics,
			       event->u.class_load.num_instance_fields,
			       event->u.class_load.instances,
			       event->u.class_load.num_methods,
			       event->u.class_load.methods,
			       event->u.class_load.class_id,
				   event->event_type);
		Unlock(methods_lock);
		return;

	case JVMPI_EVENT_CLASS_UNLOAD: 
		ClassUnloadEvent(event->u.class_unload.class_id);
		return;
	
	case JVMPI_EVENT_THREAD_START:
	case JVMPI_EVENT_THREAD_START | JVMPI_REQUESTED_EVENT:
		{
			debug("threadstart: ", event->env_id, event->u.thread_start.thread_env_id, (event->u.thread_start.thread_name != null) ? event->u.thread_start.thread_name : "null");
			ThreadInfo ti(event->u.thread_start.thread_name,
				event->u.thread_start.group_name,
				event->u.thread_start.parent_name,
				event->u.thread_start.thread_id,
				event->u.thread_start.thread_env_id);
			//cerr << event->u.thread_start.thread_env_id << endl;
			ThreadStartHandler(event->u.thread_start.thread_env_id, ti);
		}
		return;
		
	case JVMPI_EVENT_THREAD_END:
		//cerr << "remove " << event->env_id << endl;
		ThreadEndHandler(event->env_id);
		return;

	case JVMPI_EVENT_OBJECT_ALLOC:
		{
			if (event->u.obj_alloc.class_id == null)
				return;

			debug("alloc", event->env_id, event->u.obj_alloc.class_id, event->u.obj_alloc.obj_id);

			// if we haven't caught the VM loading classes, we want to see if we can
			if (!CaughtVM)
			{
				// check thread
				Lock(threads_lock);
				bool loadThread = (Threads.find(event->env_id) == Threads.end()) &&
								  (DeadThreads.find(event->env_id) == DeadThreads.end());
				Unlock(threads_lock);

				if (loadThread)
				{
					jobjectID o = CALL(GetThreadObject)(event->env_id);
					if ((o != null) && (o != event->u.obj_alloc.obj_id))
					{
						debug("in alloc, about to call RequestEvent JVMPI_EVENT_THREAD_START" , o);
						CHECKEDCALL(RequestEvent(JVMPI_EVENT_THREAD_START, o));
						debug("back from thread start");
						CaughtVM = true;
						CHECKEDCALL(DisableEvent(JVMPI_EVENT_OBJECT_ALLOC, NULL));
					}
				}
			}
			else
			{
			//	debug("CaughtVM == true!");
				// should never be reached
			}
		}
		return;

	case JVMPI_EVENT_JVM_SHUT_DOWN:
		{
		ShutdownRequested = true;
		CALL(RawMonitorNotifyAll(BackgroundMonitor));
		
		//{ for (int tdu = 0; tdu < 10; ++tdu) ThreadDump("???\n", event->env_id); }

		//Lock(output_lock);
		ShutdownCommunications();
		//Unlock(output_lock);
		
		ShutdownMonitors();
		//DeleteMethods();
		}
		return;

	case JVMPI_EVENT_JVM_INIT_DONE:
		debug("Init done!!!", event->env_id );
		CreateBackgroundThread();
		return;


	default:
		unsafe_debug("Unhandled event", event->event_type);
		return;
	} // end switch
} // end notifyEvent



